@@ -454,7 +454,7 @@ int save_database(const uint32_t remote)
   FILE *f;
 
   if(!save_file)
-    return 1;
+    return 0;
 
 #ifdef FAKEROOT_DB_PATH
   path[DB_PATH_LEN] = '\0';
@@ -464,9 +464,30 @@ int save_database(const uint32_t remote)
     roots = "/";
 #endif
 
+  do {
+    int r,fd=0;
+    struct stat s;
+    r=stat(save_file,&s);
+    if (r<0) {
+       if (errno == ENOENT)
+	  break;
+       else
+	  return EOF;
+    }
+    if (!(s.st_mode&S_IFIFO)) break;
+    fd=open(save_file,O_WRONLY|O_NONBLOCK);
+    if (fd<0) {
+      sleep(1);
+      continue;
+    }
+    close(fd);
+    break;
+  } while (1);
+
+
   f=fopen(save_file, "w");
   if(!f)
-    return 0;
+    return EOF;
 
   for (i = data_begin(); i != data_end(); i = data_node_next(i)) {
     if (i->remote != remote)
@@ -549,10 +570,10 @@ void debug_stat(const struct fakestat *st){
   fprintf(stderr,"dev:ino=(%llx:%lli), mode=0%lo, own=(%li,%li), nlink=%li, rdev=%lli\n",
 	  st->dev,
 	  st->ino,
-	  st->mode,
-	  st->uid,
-	  st->gid,
-	  st->nlink,
+	  (long)st->mode,
+	  (long)st->uid,
+	  (long)st->gid,
+	  (long)st->nlink,
 	  st->rdev);
 }
 
@@ -628,12 +649,37 @@ void process_chmod(struct fake_msg *buf){
   
   if(debug)
     fprintf(stderr,"FAKEROOT: chmod, mode=%lo\n",
-	    buf->st.mode);
+	    (long)buf->st.mode);
   
   i = data_find(&buf->st, buf->remote);
   if (i != data_end()) {
     st = data_node_get(i);
-    st->mode = (buf->st.mode&~S_IFMT) | (st->mode&S_IFMT);
+    /* Statically linked binaries can remove inodes without us knowing.
+       ldconfig is a prime offender.  Also, some packages run tests without
+       LD_PRELOAD.
+
+       While those cases can be fixed in other ways, we shouldn't continue to
+       cache stale file information.
+
+       mknod() creates a regular file, everything else should have the same
+       file type on disk and in our database.  Therefore, we check the file's
+       type first.  If we have something in our database as a device node and
+       we get a request to change it to regular file, it might be a chmod of
+       a device node that was created from within fakeroot, which is a device
+       file on disk - there's no way to distinguish.   For anything else, we
+       trust the new type and assume the inode got unlinked from something that
+       wasn't using the LD_PRELOAD library.
+    */
+
+    if ((buf->st.mode&S_IFMT) != (st->mode&S_IFMT) &&
+        ((buf->st.mode&S_IFMT) != S_IFREG || (!st->mode&(S_IFBLK|S_IFCHR)))) {
+      fprintf(stderr,"FAKEROOT: chmod mode=%lo incompatible with "
+              "existing mode=%lo\n", buf->st.mode, st->mode);
+      st->mode = buf->st.mode;
+    }
+    else{
+      st->mode = (buf->st.mode&~S_IFMT) | (st->mode&S_IFMT);
+    }
   }
   else{
     st=&buf->st;
@@ -649,7 +695,7 @@ void process_mknod(struct fake_msg *buf){
   
   if(debug)
     fprintf(stderr,"FAKEROOT: mknod, mode=%lo\n",
-	    buf->st.mode);
+	    (long)buf->st.mode);
   
   i = data_find(&buf->st, buf->remote);
   if (i != data_end()) {
@@ -742,7 +788,7 @@ void get_msg()
   int r = 0;
 
   if(debug)
-    fprintf(stderr,"FAKEROOT: msg=%i, key=%i\n",msg_get,msg_key);
+    fprintf(stderr,"FAKEROOT: msg=%i, key=%li\n",msg_get,(long)msg_key);
   do {
     r=msgrcv(msg_get,&buf,sizeof(struct fake_msg),0,0);
     if(debug)
@@ -862,12 +908,13 @@ void get_msg(const int listen_sd)
 /***********/
 
 void save(int dummy){
-  if(!save_database(0)) {
-    if(debug)
+  int savedb_state;
+  savedb_state = save_database(0);
+  if(!savedb_state) {
+    if(debug && save_file)
       fprintf(stderr, "fakeroot: saved database in %s\n", save_file);
   } else
-    if(debug)
-      fprintf(stderr, "fakeroot: database save FAILED\n");
+    fprintf(stderr, "fakeroot: database save FAILED\n");
 }
 
 #ifdef FAKEROOT_FAKENET
@@ -1048,7 +1095,7 @@ int main(int argc, char **argv){
   }
     
   if(debug)
-    fprintf(stderr,"using %i as msg key\n",msg_key);
+    fprintf(stderr,"using %li as msg key\n",(long)msg_key);
   
   msg_get=msgget(msg_key,IPC_CREAT|0600);
   msg_snd=msgget(msg_key+1,IPC_CREAT|0600);
@@ -1064,7 +1111,7 @@ int main(int argc, char **argv){
   }
 
   if(debug)
-    fprintf(stderr,"msg_key=%i\n",msg_key);
+    fprintf(stderr,"msg_key=%li\n",(long)msg_key);
 
   if(justcleanup)
     cleanup(0);
@@ -1163,12 +1210,12 @@ int main(int argc, char **argv){
 	  close(fl);
       setsid();
     } else {
-      printf("%i:%i\n",FAKE_KEY,pid);
+      printf("%li:%i\n",(long)FAKE_KEY,pid);
 
       exit(0);
     }
   } else {
-    printf("%i:%i\n",FAKE_KEY,getpid());
+    printf("%li:%i\n",(long)FAKE_KEY,getpid());
     fflush(stdout);
   }
 
