@@ -23,6 +23,29 @@
 
 #include "config.h"
 #include "communicate.h"
+
+#ifdef STAT64_SUPPORT 
+#define INT_STRUCT_STAT struct stat64
+#define INT_NEXT_STAT(a,b,c) NEXT_STAT64(a,b,c)
+#define INT_NEXT_LSTAT(a,b,c) NEXT_LSTAT64(a,b,c)
+#define INT_NEXT_FSTAT(a,b,c) NEXT_FSTAT64(a,b,c)
+#ifndef STUPID_ALPHA_HACK
+#define INT_SEND_STAT(a,b) send_stat64(a,b)
+#else
+#define INT_SEND_STAT(a,b,c) send_stat64(a,b,c)
+#endif
+#else
+#define INT_STRUCT_STAT struct stat
+#define INT_NEXT_STAT(a,b,c) NEXT_STAT(a,b,c)
+#define INT_NEXT_LSTAT(a,b,c) NEXT_LSTAT(a,b,c)
+#define INT_NEXT_FSTAT(a,b,c) NEXT_FSTAT(a,b,c)
+#ifndef STUPID_ALPHA_HACK
+#define INT_SEND_STAT(a,b) send_stat(a,b)
+#else
+#define INT_SEND_STAT(a,b,c) send_stat(a,b,c)
+#endif
+#endif
+
 #include <stdlib.h>
 #include <sys/ipc.h>
 #include <sys/msg.h>
@@ -34,6 +57,18 @@
 #include <unistd.h> 
 #include <dirent.h>
 #include <errno.h>
+#include <sys/types.h>
+#ifdef HAVE_SYS_ACL_H
+#include <sys/acl.h>
+#endif /* HAVE_SYS_ACL_H */
+
+#if !HAVE_DECL_SETENV
+extern int setenv (const char *name, const char *value, int replace);
+#endif
+#if !HAVE_DECL_UNSETENV
+extern int unsetenv (const char *name);
+#endif
+
 
 /* 
    Where are those shared libraries? 
@@ -114,7 +149,7 @@ void load_library_symbols(void){
 
   static int done=0;
   int i;
-  char* msg;
+  const char *msg;
   
   if(!done){
     for(i=0; next_wrap[i].doit; i++){
@@ -525,6 +560,27 @@ int WRAP_FSTAT FSTAT_ARG(int ver,
   return 0;
 }
 
+#ifdef HAVE_FSTATAT
+int WRAP_FSTATAT FSTATAT_ARG(int ver,
+			     int dir_fd,
+			     const char *path,
+			     struct stat *st,
+			     int flags){
+
+
+  int r;
+
+  r=NEXT_FSTATAT(ver, dir_fd, path, st, flags);
+  if(r)
+    return -1;
+#ifndef STUPID_ALPHA_HACK
+  send_get_stat(st);
+#else
+  send_get_stat(st,ver);
+#endif
+  return 0;
+}
+#endif /* HAVE_FSTATAT */
 
 #ifdef STAT64_SUPPORT
 
@@ -582,7 +638,29 @@ int WRAP_FSTAT64 FSTAT64_ARG(int ver,
   return 0;
 }
 
+#ifdef HAVE_FSTATAT
+int WRAP_FSTATAT64 FSTATAT64_ARG(int ver,
+				 int dir_fd,
+				 const char *path,
+				 struct stat64 *st,
+				 int flags){
+
+
+  int r;
+
+  r=NEXT_FSTATAT64(ver, dir_fd, path, st, flags);
+  if(r)
+    return -1;
+#ifndef STUPID_ALPHA_HACK
+  send_get_stat64(st);
+#else
+  send_get_stat64(st,ver);
 #endif
+  return 0;
+}
+#endif /* HAVE_FSTATAT */
+
+#endif /* STAT64_SUPPORT */
 
 
 
@@ -615,17 +693,17 @@ int WRAP_FSTAT64 FSTAT64_ARG(int ver,
  */
 
 int chown(const char *path, uid_t owner, gid_t group){
-  struct stat st;
+  INT_STRUCT_STAT st;
   int r=0;
 
 
 #ifdef LCHOWN_SUPPORT
   /*chown(sym-link) works on the target of the symlink if lchown is
     present and enabled.*/
-  r=NEXT_STAT(_STAT_VER, path, &st);
+  r=INT_NEXT_STAT(_STAT_VER, path, &st);
 #else
   /*chown(sym-link) works on the symlink itself, use lstat: */
-  r=NEXT_LSTAT(_STAT_VER, path, &st);
+  r=INT_NEXT_LSTAT(_STAT_VER, path, &st);
 #endif
   
   if(r)
@@ -633,9 +711,9 @@ int chown(const char *path, uid_t owner, gid_t group){
   st.st_uid=owner;
   st.st_gid=group;
 #ifndef STUPID_ALPHA_HACK
-  send_stat(&st,chown_func);
+  INT_SEND_STAT(&st,chown_func);
 #else
-  send_stat(&st,chown_func, _STAT_VER);
+  INT_SEND_STAT(&st,chown_func, _STAT_VER);
 #endif
   if(!dont_try_chown())
     r=next_lchown(path,owner,group);
@@ -701,20 +779,66 @@ int fchown(int fd, uid_t owner, gid_t group){
   return r;
 }
 
-int chmod(const char *path, mode_t mode){
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_FCHOWNAT
+int fchownat(int dir_fd, const char *path, uid_t owner, gid_t group, int flags) {
+  int r;
+  /* If AT_SYMLINK_NOFOLLOW is set in the fchownat call it should
+     be when we stat it. */
+#ifdef STAT64_SUPPORT
+  struct stat64 st;
+  r=NEXT_FSTATAT64(_STAT_VER, dir_fd, path, &st, (flags & AT_SYMLINK_NOFOLLOW));
+#else
   struct stat st;
+  r=NEXT_FSTATAT(_STAT_VER, dir_fd, path, &st, (flags & AT_SYMLINK_NOFOLLOW));
+#endif
+  
+  if(r)
+    return(r);
+  
+  st.st_uid=owner;
+  st.st_gid=group;
+#ifdef STAT64_SUPPORT
+#ifndef STUPID_ALPHA_HACK
+  send_stat64(&st,chown_func);  
+#else
+  send_stat64(&st,chown_func, _STAT_VER);  
+#endif
+#else
+#ifndef STUPID_ALPHA_HACK
+  send_stat(&st,chown_func);  
+#else
+  send_stat(&st,chown_func, _STAT_VER);  
+#endif
+#endif /* STAT64_SUPPORT */
+  
+  if(!dont_try_chown())
+    r=next_fchownat(dir_fd,path,owner,group,flags);
+  else
+    r=0;
+  
+  if(r&&(errno==EPERM))
+    r=0;
+  
+  return r;
+}
+#endif /* HAVE_FCHOWNAT */
+#endif /* HAVE_FSTATAT */
+
+int chmod(const char *path, mode_t mode){
+  INT_STRUCT_STAT st;
   int r;
 
-  r=NEXT_STAT(_STAT_VER, path, &st);
+  r=INT_NEXT_STAT(_STAT_VER, path, &st);
   if(r)
     return r;
   
   st.st_mode=(mode&ALLPERMS)|(st.st_mode&~ALLPERMS);
     
 #ifndef STUPID_ALPHA_HACK
-  send_stat(&st, chmod_func);
+  INT_SEND_STAT(&st, chmod_func);
 #else
-  send_stat(&st, chmod_func, _STAT_VER);
+  INT_SEND_STAT(&st, chmod_func, _STAT_VER);
 #endif
   
   /* if a file is unwritable, then root can still write to it
@@ -732,15 +856,55 @@ int chmod(const char *path, mode_t mode){
   r=next_chmod(path, mode);
   if(r&&(errno==EPERM))
     r=0;
+#ifdef EFTYPE		/* available under FreeBSD kernel */
+  if(r&&(errno==EFTYPE))
+    r=0;
+#endif
   return r;
 }
 
 int fchmod(int fd, mode_t mode){
+  int r;
+  INT_STRUCT_STAT st;
+
+
+  r=INT_NEXT_FSTAT(_STAT_VER, fd, &st);
+  
+  if(r)
+    return(r);
+  
+  st.st_mode=(mode&ALLPERMS)|(st.st_mode&~ALLPERMS);
+#ifndef STUPID_ALPHA_HACK
+  INT_SEND_STAT(&st,chmod_func);  
+#else
+  INT_SEND_STAT(&st,chmod_func, _STAT_VER);  
+#endif
+  
+  /* see chmod() for comment */
+  mode |= 0600;
+  if(S_ISDIR(st.st_mode))
+    mode |= 0100;
+  
+  r=next_fchmod(fd, mode);
+  if(r&&(errno==EPERM))
+    r=0;
+#ifdef EFTYPE		/* available under FreeBSD kernel */
+  if(r&&(errno==EFTYPE))
+    r=0;
+#endif
+  return r;
+}
+
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_FCHMODAT
+int fchmodat(int dir_fd, const char *path, mode_t mode, int flags) {
+/*   (int fd, mode_t mode){*/
   int r;
   struct stat st;
 
-
-  r=NEXT_FSTAT(_STAT_VER, fd, &st);
+  /* If AT_SYMLINK_NOFOLLOW is set in the fchownat call it should
+     be when we stat it. */
+  r=NEXT_FSTATAT(_STAT_VER, dir_fd, path, &st, flags & AT_SYMLINK_NOFOLLOW);
   
   if(r)
     return(r);
@@ -757,11 +921,17 @@ int fchmod(int fd, mode_t mode){
   if(S_ISDIR(st.st_mode))
     mode |= 0100;
   
-  r=next_fchmod(fd, mode);
+  r=next_fchmodat(dir_fd, path, mode, flags);
   if(r&&(errno==EPERM))
     r=0;
+#ifdef EFTYPE		/* available under FreeBSD kernel */
+  if(r&&(errno==EFTYPE))
+    r=0;
+#endif
   return r;
 }
+#endif /* HAVE_FCHMODAT */
+#endif /* HAVE_FSTATAT */
 
 int WRAP_MKNOD MKNOD_ARG(int ver UNUSED,
 			 const char *pathname, 
@@ -802,8 +972,54 @@ int WRAP_MKNOD MKNOD_ARG(int ver UNUSED,
   return 0;
 }
 
-int mkdir(const char *path, mode_t mode){
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_MKNODAT
+int WRAP_MKNODAT MKNODAT_ARG(int ver UNUSED,
+			     int dir_fd,
+			     const char *pathname, 
+			     mode_t mode, dev_t XMKNODAT_FIFTH_ARG dev)
+{
   struct stat st;
+  mode_t old_mask=umask(022);
+  int fd,r;
+
+  umask(old_mask);
+  
+  /*Don't bother to mknod the file, that probably doesn't work.
+    just create it as normal file, and leave the permissions
+    to the fakemode.*/
+
+  fd=openat(dir_fd, pathname, O_WRONLY|O_CREAT|O_TRUNC, 00644);
+
+  if(fd==-1)
+    return -1;
+  
+  close(fd);
+  /* get the inode, to communicate with faked */
+
+  /* The only known flag is AT_SYMLINK_NOFOLLOW and
+     we don't want that here. */
+  r=NEXT_FSTATAT(_STAT_VER, dir_fd, pathname, &st, 0);
+
+  if(r)
+    return -1;
+  
+  st.st_mode= mode & ~old_mask;
+  st.st_rdev= XMKNODAT_FIFTH_ARG dev;
+  
+#ifndef STUPID_ALPHA_HACK
+  send_stat(&st,mknod_func);
+#else
+  send_stat(&st,mknod_func, _STAT_VER);
+#endif
+    
+  return 0;
+}
+#endif /* HAVE_MKNODAT */
+#endif /* HAVE_FSTATAT */
+
+int mkdir(const char *path, mode_t mode){
+  INT_STRUCT_STAT st;
   int r;
   mode_t old_mask=umask(022);
 
@@ -818,13 +1034,47 @@ int mkdir(const char *path, mode_t mode){
   /* mode|0700: see comment in the chown() function above */
   if(r)
     return -1;
-  r=NEXT_STAT(_STAT_VER, path, &st);
+  r=INT_NEXT_STAT(_STAT_VER, path, &st);
   
   if(r)
     return -1;
   
   st.st_mode=(mode&~old_mask&ALLPERMS)|(st.st_mode&~ALLPERMS)|S_IFDIR;
 
+#ifndef STUPID_ALPHA_HACK
+  INT_SEND_STAT(&st, chmod_func);
+#else
+  INT_SEND_STAT(&st, chmod_func, _STAT_VER);
+#endif
+
+  return 0;
+}
+
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_MKDIRAT
+int mkdirat(int dir_fd, const char *path, mode_t mode){
+  struct stat st;
+  int r;
+  mode_t old_mask=umask(022);
+
+  umask(old_mask);
+
+
+  /* we need to tell the fake deamon the real mode. In order
+     to communicate with faked we need a struct stat, so we now
+     do a stat of the new directory (just for the inode/dev) */
+
+  r=next_mkdirat(dir_fd, path, mode|0700); 
+  /* mode|0700: see comment in the chown() function above */
+  if(r)
+    return -1;
+  r=NEXT_FSTATAT(_STAT_VER, dir_fd, path, &st, 0);
+
+  if(r)
+    return -1;
+  
+  st.st_mode=(mode&~old_mask&ALLPERMS)|(st.st_mode&~ALLPERMS)|S_IFDIR;
+
 #ifndef STUPID_ALPHA_HACK
   send_stat(&st, chmod_func);
 #else
@@ -833,6 +1083,8 @@ int mkdir(const char *path, mode_t mode){
 
   return 0;
 }
+#endif /* HAVE_MKDIRAT */
+#endif /* HAVE_FSTATAT */
 
 /* 
    The remove funtions: unlink, rmdir, rename.
@@ -851,13 +1103,10 @@ int mkdir(const char *path, mode_t mode){
 
 int unlink(const char *pathname){
   int r;
-#ifdef STAT64_SUPPORT
-  struct stat64 st;
-  r=NEXT_LSTAT64(_STAT_VER, pathname, &st);
-#else
-  struct stat st;
-  r=NEXT_LSTAT(_STAT_VER, pathname, &st);
-#endif
+  INT_STRUCT_STAT st;
+
+
+  r=INT_NEXT_LSTAT(_STAT_VER, pathname, &st);
   if(r)
     return -1;
 
@@ -866,6 +1115,34 @@ int unlink(const char *pathname){
   if(r)
     return -1;
   
+#ifndef STUPID_ALPHA_HACK
+  INT_SEND_STAT(&st, unlink_func);
+#else
+  INT_SEND_STAT(&st, unlink_func, _STAT_VER);
+#endif
+  
+  return 0;
+}
+
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_UNLINKAT
+int unlinkat(int dir_fd, const char *pathname, int flags){
+  int r;
+#ifdef STAT64_SUPPORT
+  struct stat64 st;
+  r=NEXT_FSTATAT64(_STAT_VER, dir_fd, pathname, &st, (flags&~AT_REMOVEDIR) | AT_SYMLINK_NOFOLLOW);
+#else
+  struct stat st;
+  r=NEXT_FSTATAT(_STAT_VER, dir_fd, pathname, &st, (flags&~AT_REMOVEDIR) | AT_SYMLINK_NOFOLLOW);
+#endif
+  if(r)
+    return -1;
+
+  r=next_unlinkat(dir_fd, pathname, flags);
+
+  if(r)
+    return -1;
+  
 #ifdef STAT64_SUPPORT
 #ifndef STUPID_ALPHA_HACK
   send_stat64(&st, unlink_func);
@@ -878,6 +1155,8 @@ int unlink(const char *pathname){
   
   return 0;
 }
+#endif /* HAVE_UNLINKAT */
+#endif /* HAVE_FSTATAT */
 
 /*
   See the `remove funtions:' comments above for more info on
@@ -885,9 +1164,9 @@ int unlink(const char *pathname){
 */
 int rmdir(const char *pathname){
   int r;
-  struct stat st;
+  INT_STRUCT_STAT st;
 
-  r=NEXT_LSTAT(_STAT_VER, pathname, &st);
+  r=INT_NEXT_LSTAT(_STAT_VER, pathname, &st);
   if(r)
     return -1;
   r=next_rmdir(pathname);  
@@ -895,9 +1174,9 @@ int rmdir(const char *pathname){
     return -1;
 
 #ifndef STUPID_ALPHA_HACK
-  send_stat(&st,unlink_func);  
+  INT_SEND_STAT(&st,unlink_func);  
 #else
-  send_stat(&st,unlink_func, _STAT_VER);  
+  INT_SEND_STAT(&st,unlink_func, _STAT_VER);  
 #endif
 
   return 0;
@@ -909,18 +1188,18 @@ int rmdir(const char *pathname){
 */
 int remove(const char *pathname){
   int r;
-  struct stat st;
+  INT_STRUCT_STAT st;
 
-  r=NEXT_LSTAT(_STAT_VER, pathname, &st);
+  r=INT_NEXT_LSTAT(_STAT_VER, pathname, &st);
   if(r)
     return -1;
   r=next_remove(pathname);  
   if(r)
     return -1;
 #ifndef STUPID_ALPHA_HACK
-  send_stat(&st,unlink_func);  
+  INT_SEND_STAT(&st,unlink_func);  
 #else
-  send_stat(&st,unlink_func, _STAT_VER);  
+  INT_SEND_STAT(&st,unlink_func, _STAT_VER);  
 #endif
   
   return r;
@@ -937,6 +1216,34 @@ int remove(const char *pathname){
 
 int rename(const char *oldpath, const char *newpath){
   int r,s;
+  INT_STRUCT_STAT st;     
+
+  /* If newpath points to an existing file, that file will be 
+     unlinked.   Make sure we tell the faked daemon about this! */
+
+  /* we need the st_new struct in order to inform faked about the
+     (possible) unlink of the file */
+
+  r=INT_NEXT_LSTAT(_STAT_VER, newpath, &st);
+
+  s=next_rename(oldpath, newpath);
+  if(s)
+    return -1;
+  if(!r)
+#ifndef STUPID_ALPHA_HACK
+    INT_SEND_STAT(&st,unlink_func);
+#else
+    INT_SEND_STAT(&st,unlink_func, _STAT_VER);
+#endif
+
+  return 0;
+}
+
+#ifdef HAVE_FSTATAT
+#ifdef HAVE_RENAMEAT
+int renameat(int olddir_fd, const char *oldpath,
+             int newdir_fd, const char *newpath){
+  int r,s;
   struct stat st;     
 
   /* If newpath points to an existing file, that file will be 
@@ -945,9 +1252,9 @@ int rename(const char *oldpath, const char *newpath){
   /* we need the st_new struct in order to inform faked about the
      (possible) unlink of the file */
 
-  r=NEXT_LSTAT(_STAT_VER, newpath, &st);
+  r=NEXT_FSTATAT(_STAT_VER, newdir_fd, newpath, &st, AT_SYMLINK_NOFOLLOW);
 
-  s=next_rename(oldpath, newpath);
+  s=next_renameat(olddir_fd, oldpath, newdir_fd, newpath);
   if(s)
     return -1;
   if(!r)
@@ -959,6 +1266,9 @@ int rename(const char *oldpath, const char *newpath){
 
   return 0;
 }
+#endif /* HAVE_RENAMEAT */
+#endif /* HAVE_FSTATAT */
+
 
 #ifdef FAKEROOT_FAKENET
 pid_t fork(void)
@@ -1167,3 +1477,15 @@ int fakeroot_isdisabled(void)
 {
   return fakeroot_disabled;
 }
+
+#ifdef HAVE_SYS_ACL_H
+int acl_set_fd(int fd, acl_t acl) {
+  errno = ENOTSUP;
+  return -1;
+}
+
+int acl_set_file(const char *path_p, acl_type_t type, acl_t acl) {
+  errno = ENOTSUP;
+  return -1;
+}
+#endif /* HAVE_SYS_ACL_H */
